import numpy as np
import pandas as pd
import CoolProp as cp
import datetime
import os
import yaml
from .. import pysolver_view as psv
from .. import utilities as utils
from . import geometry_model as geom
from . import flow_model as flow
from .. import properties as props
from . import performance_analysis as pa


CONSTRAINTS = ["mass_flow_rate", "interstage_flaring"]
AVAILABLE_INEQ_CONSTRAINTS = ["mass_flow_rate", "interstage_flaring"]
OBJECTIVE_FUNCTIONS = ["efficiency_ts", "none"]
RADIUS_TYPE = ["constant_mean", "constant_hub", "constant_tip"]
ANGLE_KEYS = ["leading_edge_angle", "gauging_angle"]
INDEPENDENT_VARIABLES = [
    "v_in",
    "w_out",
    "s_out",
    "beta_out",
    "v*_in",
    "beta*_throat",
    "w*_throat",
    "s*_throat",
]
INDEXED_VARIABLES = [
    "hub_tip_ratio_in",
    "hub_tip_ratio_out",
    "aspect_ratio",
    "pitch_chord_ratio",
    "trailing_edge_thickness_opening_ratio",
    "leading_edge_angle",
    "gauging_angle",
    "throat_location_fraction",
    "leading_edge_diameter",
    "leading_edge_wedge_angle",
    "tip_clearance",
    "cascade_type",
]
VARIABLES = (
    INDEXED_VARIABLES + ["specific_speed", "blade_jet_ratio"] + INDEPENDENT_VARIABLES
)


def compute_optimal_turbine(
    config,
    out_filename=None,
    out_dir="output",
    export_results=True,
):
    r"""
    Calculate the optimal turbine configuration based on the specified optimization problem.

    The function checks the configuration file before performing design optimization.

    Parameters
    ----------
    config : dict
        A dictionary containing necessary configuration options for computing optimal turbine.
    initial_guess : None or dict
        The initial guess of the design optimization. If None, a defualt initial guess is provided. If given, he initial guess must correspond
        with the given set of design variables given in the configuration dictionary.

    Returns
    -------
    solution : object
        The solution object containing the results of the design optimization.
    """

    # Initialize problem object
    problem = CascadesOptimizationProblem(config)

    # Perform initial function call to initialize problem
    # This populates the arrays of equality and inequality constraints
    # TODO: it might be more intuitive to create a new method called initialize_problem() that generates the initial guess and evaluates the fitness() function with it
    problem.fitness(problem.initial_guess)

    # Load solver configuration
    solver_config = config["design_optimization"]["solver_options"]

    # Initialize solver object using keyword-argument dictionary unpacking
    solver = psv.OptimizationSolver(problem, **solver_config)

    # Solve optimization problem for initial guess x0
    solver.solve(problem.initial_guess)

    dfs = {
        "operation point": pd.DataFrame(
            {key: pd.Series(val) for key, val in problem.boundary_conditions.items()}
        ),
        "overall": pd.DataFrame(problem.results["overall"], index=[0]),
        "plane": problem.results["plane"],
        "cascade": problem.results["cascade"],
        "stage": problem.results["stage"],
        "geometry": pd.DataFrame(
            {key: pd.Series(val) for key, val in problem.geometry.items()}
        ),
        "solver": pd.DataFrame(
            {
                "completed": pd.Series(True, index=[0]),
                "success": pd.Series(solver.success, index=[0]),
                "message": pd.Series(solver.message, index=[0]),
                "grad_count": solver.convergence_history["grad_count"],
                "func_count": solver.convergence_history["func_count"],
                "func_count_total": solver.convergence_history["func_count_total"],
                "objective_value": solver.convergence_history["objective_value"],
                "constraint_violation": solver.convergence_history[
                    "constraint_violation"
                ],
                "norm_step": solver.convergence_history["norm_step"],
            },
            index=range(len(solver.convergence_history["grad_count"])),
        ),
    }

    if export_results:
        # Create a directory to save simulation results
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        # Define filename with unique date-time identifier
        if out_filename == None:
            current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            out_filename = f"design_optimization_{current_time}"

        # Export simulation configuration as YAML file
        config_data = {k: v for k, v in config.items() if v}  # Filter empty entries
        config_data = utils.convert_numpy_to_python(config_data, precision=12)
        config_file = os.path.join(out_dir, f"{out_filename}.yaml")
        with open(config_file, "w") as file:
            yaml.dump(config_data, file, default_flow_style=False, sort_keys=False)

        # Export optimal turbine in excel file
        filepath = os.path.join(out_dir, f"{out_filename}.xlsx")
        with pd.ExcelWriter(filepath, engine="openpyxl") as writer:
            for sheet_name, df in dfs.items():
                df.to_excel(writer, sheet_name=sheet_name, index=True)

    return solver


class CascadesOptimizationProblem(psv.OptimizationProblem):
    """
    A class representing a turbine design optimization problem.

    This class is designed for solving design optimization problems of axial turbines. It is initialized by providing a
    dictionary containing information on design variables, objective function, constraints and bounds.

    Parameters
    ----------
    config : dictionary
        A dictionary containing necessary configuration options for computing optimal turbine.

    Attributes
    ----------
    fluid : FluidCoolProp_2Phase
        An instance of the FluidCoolProp_2Phase class representing the fluid properties.
    results : dict
        A dictionary to store results.
    boundary_conditions : dict
        A dictionary containing boundary condition data.
    geometry : dict
        A dictionary containing geometry-related data.
    model_options : dict
        A dictionary containing options related to the analysis model.
    reference_values : dict
        A dictionary containing reference values for calculations.
    design_variables_keys : list
        A list of strings defining the selected design variables.
    objective_function : str
        A string defining the selected objective function for the design optimization problem.
    eq_constraints : list
        A list of strings containing the selected equaility constraints for the design optimization problem.
    ineq_constraints : list
        A list of strings containing the selected inequaility constraints for the design optimization problem.
    bounds : list
        A list of tuples on the form (lower bound, upper bound) defining the bounds for the design optimization problem.
    radius_type : str
        A string deciding what type of turbine geometry should be condidered (constant mean, hub or tip radius).
    vars_scaled : dict
        A dict containing the scaled flow design variables used to evaluate turbine performance.


    Methods
    -------
    fitness(x)
        Evaluate the objective function and constraints at a given point `x`.
    get_bounds()
        Provide the bounds for the design optimization problem.
    get_n_eq()
        Get the number of equality constraints.
    get_n_ineq()
        Get the number of oinequality constraints.
    get_omega(specific_speed, mass_flow, h0_in, d_is, h_is)
        Convert specific speed to actual angular speed
    extend_variables(variables)
        Convert a dict containing lists or arrays to a dictionaries of only one element.
    get_given_bounds(design_variable)
        Retrieves the lower and upper bounds for the given design variables.
    get_constraints(self, constraints):
        Converts a list of constraints to equality and inequality constraints.
    update_boundary_conditions(design_point)
        Update the boundary conditions of the turbine analysis problem with the design point of the design optimization problem.
    convert_performance_analysis_results(performance_problem)
        Generate a feasable set of initial guess from a solved performance analysis problem object.

    """

    def __init__(self, config):
        r"""
        Initialize a CascadesOptimizationProblem.

        Parameters
        ----------
        config : dict
            A dictionary containing case-specific data.

        """

        # Get list of design variables
        self.obj_func = config["design_optimization"]["objective_function"]
        self.radius_type = config["design_optimization"]["radius_type"]
        self.eq_constraints, self.ineq_constraints = self.get_constraints(
            config["design_optimization"]["constraints"]
        )

        # Update design point
        self.update_boundary_conditions(config["operation_points"])

        # Initialize model options
        self.model_options = config["simulation_options"]

        # Adjust given variables according to choking model (remove excess variables)
        variables = config["design_optimization"]["variables"]
        if self.model_options["choking_model"] == "evaluate_cascade_throat":
            variables = {
                key: var
                for key, var in variables.items()
                if not key.startswith("v*_in")
            }
        elif self.model_options["choking_model"] == "evaluate_cascade_critical":
            variables = {
                key: var
                for key, var in variables.items()
                if not key.startswith("beta*_throat")
            }
        elif (
            self.model_options["choking_model"] == "evaluate_cascade_isentropic_throat"
        ):
            variables = {
                key: var
                for key, var in variables.items()
                if not (
                    key.startswith("v*_in")
                    or key.startswith("s*_throat")
                    or key.startswith("beta*_throat")
                )
            }

        # Separate fixed variables and design variables
        fixed_params = {}
        design_variables = {}
        for key, value in variables.items():
            if value["lower_bound"] is not None:
                design_variables[key] = value
            else:
                fixed_params[key] = value

        # Initialize initial guess
        initial_guess = self.extend_variables(design_variables)
        self.design_variables_keys = initial_guess.keys()
        self.initial_guess = np.array(list(initial_guess.values()))

        # Extend fixed variables
        self.fixed_params = self.extend_variables(fixed_params)

        # Initialize bounds
        lb, ub = self.get_given_bounds(design_variables)
        self.bounds = (lb, ub)

        # Adjust set of independent variables according to choking model
        self.independent_variables = [
            key
            for key in initial_guess.keys()
            if any(key.startswith(var) for var in INDEPENDENT_VARIABLES)
        ]
        if self.model_options["choking_model"] == "evaluate_cascade_throat":
            self.independent_variables = [
                var for var in self.independent_variables if not var.startswith("v*_in")
            ]
        elif self.model_options["choking_model"] == "evaluate_cascade_critical":
            self.independent_variables = [
                var
                for var in self.independent_variables
                if not var.startswith("beta*_throat")
            ]
        elif (
            self.model_options["choking_model"] == "evaluate_cascade_isentropic_throat"
        ):
            self.independent_variables = [
                var
                for var in self.independent_variables
                if not (
                    var.startswith("v*_in")
                    or var.startswith("s*_throat")
                    or var.startswith("beta*_throat")
                )
            ]

    def extend_variables(self, variables):
        """
        Index the design variables for each cascade for the default initial guess.

        The design variables are given without indexing the cascades. For example, the aspect ratio is simply provided by aspect_ratio.
        This function extend the design_variable dictionary such that for each cascade specific design variable, one item is given for each cascade.
        for example `aspect_ratio_1` for first cascade and etc.

        Parameters
        ----------
        variables : dict
            A dictionary where each key maps to a dictionary containing at least a
            "value" key. The "value" can be a list, numpy array, or float.

        Returns
        -------
        dict
            A dictionary where each list or numpy array in the original dictionary
            is expanded into individual elements with keys formatted as "key_index".
            Float values are kept as is.

        """

        variables_extended = {}
        for key in variables.keys():
            if isinstance(variables[key]["value"], (list, np.ndarray)):
                for i in range(len(variables[key]["value"])):
                    variables_extended[f"{key}_{i+1}"] = variables[key]["value"][i]
            elif isinstance(variables[key]["value"], float):
                variables_extended[key] = variables[key]["value"]

        return variables_extended

    def get_given_bounds(self, design_variables):
        """
        Retrieves the lower and upper bounds for the given design variables.

        Parameters
        ----------
        design_variables : dict
            A dictionary where each key maps to a dictionary containing a
            "value" key, a "lower_bound" key, and an "upper_bound" key. The values
            can be a list, numpy array, or float.

        Returns
        -------
        lb : list
            A list of lower bounds for the design variables.
        ub : list
            A list of upper bounds for the design variables.

        """

        lb = []
        ub = []
        for key in design_variables.keys():
            if isinstance(design_variables[key]["value"], list):
                lb += design_variables[key]["lower_bound"]
                ub += design_variables[key]["upper_bound"]
            elif isinstance(design_variables[key]["value"], np.ndarray):
                lb += list(design_variables[key]["lower_bound"])
                ub += list(design_variables[key]["upper_bound"])
            elif isinstance(design_variables[key]["value"], float):
                lb += [design_variables[key]["lower_bound"]]
                ub += [design_variables[key]["upper_bound"]]

        return lb, ub

    def fitness(self, x):
        r"""
        Evaluate the objective function and constraints at a given pint `x`.

        Parameters
        ----------
        x : array-like
            Vector of design variables.

        Returns
        -------
        numpy.ndarray
            An array containg the value of the objective function and constraints.
        """

        # Rename reference values
        h0_in = self.boundary_conditions["h0_in"]
        mass_flow = self.reference_values["mass_flow_ref"]
        h_is = self.reference_values["h_out_s"]
        d_is = self.reference_values["d_out_s"]
        v0 = self.reference_values["v0"]
        angle_range = self.reference_values["angle_range"]
        angle_min = self.reference_values["angle_min"]

        # Structure design variables to dictionary (Assume set of design variables)
        design_variables = dict(zip(self.design_variables_keys, x))

        # Construct array with independent variables
        self.vars_scaled = {
            key: design_variables[key] for key in self.independent_variables
        }

        # Contruct variables
        variables = {**design_variables, **self.fixed_params}

        # Calculate angular speed
        specific_speed = variables["specific_speed"]
        self.boundary_conditions["omega"] = self.get_omega(
            specific_speed, mass_flow, h0_in, d_is, h_is
        )

        # Calculate mean radius
        blade_jet_ratio = variables["blade_jet_ratio"]
        blade_speed = blade_jet_ratio * v0
        self.geometry = {}
        self.geometry["radius"] = blade_speed / self.boundary_conditions["omega"]

        # Assign geometry design variables to geometry attribute
        for key in INDEXED_VARIABLES:
            self.geometry[key] = np.array(
                [v for k, v in variables.items() if k.startswith(key)]
            )
            if key in ANGLE_KEYS:
                self.geometry[key] = self.geometry[key] * angle_range + angle_min

        self.geometry = geom.prepare_geometry(self.geometry, self.radius_type)
        self.geometry = geom.calculate_full_geometry(self.geometry)

        # Evaluate turbine model
        self.results = flow.evaluate_axial_turbine(
            self.vars_scaled,
            self.boundary_conditions,
            self.geometry,
            self.fluid,
            self.model_options,
            self.reference_values,
        )

        # Evaluate available objecitve functions
        objective_functions = {
            "none": 0,
            "efficiency_ts": -self.results["overall"]["efficiency_ts"].values[0] / 10,
        }
        self.f = objective_functions[self.obj_func]

        # Evaluate available equality constraints
        available_constraints = {
            "mass_flow_rate": self.results["overall"]["mass_flow_rate"].values[0],
            "interstage_flaring": self.geometry["A_in"][1:]
            / self.geometry["A_out"][0:-1],
            "flaring": self.geometry["flaring_angle"],
        }
        self.c_eq = np.array(list(self.results["residuals"].values()))
        self.c_ineq = np.array([])
        for constraint in self.eq_constraints:
            self.c_eq = np.append(
                self.c_eq,
                (available_constraints[constraint["variable"]] - constraint["value"])
                / constraint["scale"],
            )

        for constraint in self.ineq_constraints:
            self.c_ineq = np.append(
                self.c_ineq,
                (constraint["value"] - available_constraints[constraint["variable"]])
                / constraint["scale"],
            )

        objective_and_constraints = psv.combine_objective_and_constraints(
            self.f, self.c_eq, self.c_ineq
        )

        return objective_and_constraints

    def get_bounds(self):
        r"""
        Provide the bounds for the design optimization problem.

        Returns
        -------
        list
            List of toutples containg the lower and upper bound for each design variable.
        """

        return self.bounds

    def get_nec(self):
        r"""
        Get the number of equality constraints.

        Returns
        -------
        int
            Number of equality constraints.

        """

        return psv.count_constraints(self.c_eq)

    def get_nic(self):
        r"""
        Get the number of inequality constraints.

        Returns
        -------
        int
            Number of inequality constraints.

        """
        return psv.count_constraints(self.c_ineq)

    def get_omega(self, specific_speed, mass_flow, h0_in, d_is, h_is):
        r"""
        Convert specific speed to actual angular speed

        Parameters
        ----------
        specific_speed : float
            Given specific speed.
        mass_flow : float
            Given mass flow rate.
        h0_in : float
            Turbine inlet stagnation enthalpy.
        d_is : float
            Turbine exit density for an isentropic expansion.
        h_is : float
            Turbine exit enthalpy for an isentropic expansion.

        Returns
        -------
        float
            Actual angular speed.

        """

        return specific_speed * (h0_in - h_is) ** (3 / 4) / ((mass_flow / d_is) ** 0.5)

    def update_boundary_conditions(self, design_point):
        """
        Update the boundary conditions of the turbine analysis problem with the design point of the design optimization problem.

        This method updates the boundary conditions attributes used to evaluate the turbine performance.
        It also initializes a Fluid object using the 'fluid_name' specified in the operation point.
        The method computes additional properties and reference values like stagnation properties at
        the inlet, exit static properties, spouting velocity, and reference mass flow rate.
        These are stored in the object's internal state for further use in calculations.

        Parameters
        ----------
        design_point : dict
            A dictionary containing the boundary conditions defining the operation point. It must include the following keys:

            - `fluid_name` (str) : The name of the fluid to be used in the Fluid object.
            - `T0_in` (float): The inlet temperature (in Kelvin).
            - `p0_in` (float): The inlet pressure (in Pascals).
            - `p_out` (float): The outlet pressure (in Pascals).
            - `omega` (float): The rotational speed (in rad/s).
            - `alpha_in` (float): The inlet flow angle (in degrees).

        Returns
        -------
        None
            This method does not return a value but updates the attributes of the object.

        """

        # Define current operating point
        self.boundary_conditions = design_point

        # Initialize fluid object
        self.fluid = props.Fluid(design_point["fluid_name"])

        # Rename variables
        p0_in = design_point["p0_in"]
        T0_in = design_point["T0_in"]
        p_out = design_point["p_out"]

        # Compute stagnation properties at inlet
        state_in_stag = self.fluid.get_props(cp.PT_INPUTS, p0_in, T0_in)
        h0_in = state_in_stag["h"]
        s_in = state_in_stag["s"]

        # Store the inlet stagnation (h,s) for the first stage
        # TODO: Improve logic of implementation?
        self.boundary_conditions["h0_in"] = h0_in
        self.boundary_conditions["s_in"] = s_in

        # Calculate exit static properties for a isentropic expansion
        state_out_s = self.fluid.get_props(cp.PSmass_INPUTS, p_out, s_in)
        h_isentropic = state_out_s["h"]
        d_isentropic = state_out_s["d"]

        # Calculate exit static properties for a isenthalpic expansion
        state_out_h = self.fluid.get_props(cp.HmassP_INPUTS, h0_in, p_out)
        s_isenthalpic = state_out_h["s"]
        # Calculate spouting velocity
        v0 = np.sqrt(2 * (h0_in - h_isentropic))

        mass_flow_rate = None  # Initialize mass_flow_rate to None
        # Try to find mass_flow_rate from eq_constraints
        for constraint in self.eq_constraints:
            if constraint.get("variable") == "mass_flow_rate":
                mass_flow_rate = constraint.get("value")
                break

        # If mass_flow_rate not found, calculate using default formula
        if mass_flow_rate is None:
            mass_flow_rate = v0 * d_isentropic

        # Define reference_values
        self.reference_values = {
            "s_range": s_isenthalpic - s_in,
            "s_min": s_in,
            "v0": v0,
            "h_out_s": h_isentropic,
            "d_out_s": d_isentropic,
            "mass_flow_ref": mass_flow_rate,
            "angle_range": 180,
            "angle_min": -90,
        }

        return

    def convert_performance_analysis_results(self, performance_problem):
        r"""
        Generate a feasable set of initial guess from a solved performance analysis problem object.

        The function adopt the vars_scaled attribute from the solved performance analysis problem, the geometry and
        angular speed to generate an initial guess of correct form to be given to the design optimization problem.

        Returns
        -------
        dict
            Feasable initial guess.
        """

        design_variables = self.design_variables_keys
        radius_type = self.radius_type
        mass_flow = self.reference_values["mass_flow_ref"]

        geometry = performance_problem.geometry
        overall = performance_problem.results["overall"]
        vars_scaled = performance_problem.vars_scaled

        angle_min = self.reference_values["angle_min"]
        angle_range = self.reference_values["angle_range"]
        d_out_s = self.reference_values["d_out_s"]
        h0_in = self.boundary_conditions["h0_in"]
        h_out_s = self.reference_values["h_out_s"]

        initial_guess = {}
        for design_variable in set(design_variables) - {
            "specific_speed",
            "blade_jet_ratio",
        }:
            if "angle" in design_variable:
                added_dict = {
                    design_variable
                    + f"_{i+1}": (geometry[design_variable][i] - angle_min)
                    / angle_range
                    for i in range(len(geometry[design_variable]))
                }
            else:
                added_dict = {
                    design_variable + f"_{i+1}": geometry[design_variable][i]
                    for i in range(len(geometry[design_variable]))
                }
            initial_guess = {**initial_guess, **added_dict}

        if "blade_jet_ratio" in design_variables:
            if radius_type == "constant_mean":
                initial_guess["blade_jet_ratio"] = overall[
                    "blade_jet_ratio_mean"
                ].values[0]
            elif radius_type == "constant_hub":
                initial_guess["blade_jet_ratio"] = overall[
                    "blade_jet_ratio_hub"
                ].values[0]
            elif radius_type == "constant_mean":
                initial_guess["blade_jet_ratio"] = overall[
                    "blade_jet_ratio_tip"
                ].values[0]

        if "specific_speed" in design_variables:
            angular_speed = overall["angular_speed"].values[0]
            initial_guess["specific_speed"] = (
                angular_speed
                * (mass_flow / d_out_s) ** 0.5
                / ((h0_in - h_out_s) ** 0.75)
            )

        return {**vars_scaled, **initial_guess}

    def get_constraints(self, constraints):
        """
        Converts a list of constraints to equality and inequality constraints.

        Parameters:
        -----------
        constraints : list of dict
            List of constraints where each constraint is represented as a dictionary
            with keys "type", "variable", "value", and "normalize". "type" represents
            the type of constraint ('=', '<', or '>'). "variable" is the variable
            involved in the constraint. "value" is the value of the constraint. "normalize"
            specifies whether to normalize the constraint value.

        Returns:
        --------
        list of dict
            List of equality constraints, each represented as a dictionary with keys
            "variable", "value", and "scale". "variable" is the variable involved
            in the constraint. "value" is the value of the constraint. "scale" is
            the scaling factor applied to the value.
        list of dict
            List of inequality constraints, each represented as a dictionary with keys
            "variable", "value", and "scale". "variable" is the variable involved
            in the constraint. "value" is the value of the constraint. "scale" is
            the scaling factor applied to the value.
        """

        eq_constraints = []
        ineq_constraints = []
        for constraint in constraints.keys():
            if constraints[constraint]["type"] == "=":
                if constraints[constraint]["normalize"]:
                    eq_constraints += [
                        {
                            "variable": constraint,
                            "value": constraints[constraint]["value"],
                            "scale": constraints[constraint]["value"],
                        }
                    ]
                else:
                    eq_constraints += [
                        {
                            "variable": constraint,
                            "value": constraints[constraint]["value"],
                            "scale": 1,
                        }
                    ]

            elif constraints[constraint]["type"] == "<":
                if constraints[constraint]["normalize"]:
                    ineq_constraints += [
                        {
                            "variable": constraint,
                            "value": constraints[constraint]["value"],
                            "scale": abs(constraints[constraint]["value"]),
                        }
                    ]
                else:
                    ineq_constraints += [
                        {
                            "variable": constraint,
                            "value": constraints[constraint]["value"],
                            "scale": 1,
                        }
                    ]

            elif constraints[constraint]["type"] == ">":
                if constraints[constraint]["normalize"]:
                    ineq_constraints += [
                        {
                            "variable": constraint,
                            "value": constraints[constraint]["value"],
                            "scale": -1 * abs(constraints[constraint]["value"]),
                        }
                    ]
                else:
                    ineq_constraints += [
                        {
                            "variable": constraint,
                            "value": constraints[constraint]["value"],
                            "scale": -1,
                        }
                    ]
        return eq_constraints, ineq_constraints



import numpy as np

from scipy.integrate import solve_ivp

# from .. import utilities
from .. import fluid_properties as props


def heat_exchanger(
    fluid_hot,
    h_in_hot,
    h_out_hot,
    p_in_hot,
    p_out_hot,
    fluid_cold,
    h_in_cold,
    h_out_cold,
    p_in_cold,
    p_out_cold,
    num_steps=50,
    counter_current=True,
):
    # Evaluate properties on the hot side
    hot_side = heat_transfer_process(
        fluid=fluid_hot,
        h_1=h_in_hot,
        h_2=h_out_hot,
        p_1=p_in_hot,
        p_2=p_out_hot,
        num_steps=num_steps,
    )

    # Evaluate properties in the cold side
    cold_side = heat_transfer_process(
        fluid=fluid_cold,
        h_1=h_in_cold,
        h_2=h_out_cold,
        p_1=p_in_cold,
        p_2=p_out_cold,
        num_steps=num_steps,
    )

    # Sort values for temperature difference calculation
    if counter_current:
        for key, value in hot_side["states"].items():
            hot_side["states"][key] = np.flip(value)

    # Compute temperature difference
    dT = hot_side["states"]["T"] - cold_side["states"]["T"]

    # Compute mass flow ratio
    # Use 1.0 for cases with no heat exchange (i.e., no recuperator)
    dh_hot = hot_side["state_in"].h - hot_side["state_out"].h
    dh_cold = cold_side["state_out"].h - cold_side["state_in"].h
    mass_ratio = 1.00 if (dh_cold == 0 or dh_hot == 0) else dh_cold / dh_hot

    # Create result dictionary
    result = {
        "type": "heat_exchanger",
        "hot_side": hot_side,
        "cold_side": cold_side,
        "q_hot_side": dh_hot,
        "q_cold_side": dh_cold,
        "temperature_difference": dT,
        "mass_flow_ratio": mass_ratio,
    }

    return result


def heat_transfer_process(fluid, h_1, p_1, h_2, p_2, num_steps=25):
    # Generate linearly spaced arrays for pressure and enthalpy
    p_array = np.linspace(p_1, p_2, num_steps)
    h_array = np.linspace(h_1, h_2, num_steps)

    # Initialize lists to store states for hot and cold sides
    states = []

    # Calculate states for hot side
    for p, h in zip(p_array, h_array):
        states.append(fluid.set_state(props.HmassP_INPUTS, h, p))

    # Store inlet and outlet states
    state_in = states[0]
    state_out = states[-1]

    # Create result dictionary
    result = {
        "states": props.states_to_dict(states),
        "fluid_name": fluid.name,
        "state_in": state_in,
        "state_out": state_out,
        "mass_flow": np.nan,
        "color": "black",
    }

    return result


def compression_process(
    fluid, h_in, p_in, p_out, efficiency, efficiency_type="isentropic", num_steps=10
):
    """
    Calculate properties along a compression process defined by a isentropic or polytropic efficiency

    Parameters
    ----------
    fluid : Fluid
        The fluid object used to evaluate thermodynamic properties
    h_in : float
        Enthalpy at the start of the compression process.
    p_in : float
        Pressure at the start of the compression process.
    p_out : float
        Pressure at the end of the compression process.
    efficiency : float
        The efficiency of the compression process.
    efficiency_type : str, optional
        The type of efficiency to be used in the process ('isentropic' or 'polytropic'). Default is 'isentropic'.
    num_steps : int, optional
        The number of steps for the polytropic process calculation. Default is 50.

    Returns
    -------
    tuple
        Tuple containing (state_out, states) where states is a list with all intermediate states

    Raises
    ------
    ValueError
        If an invalid 'efficiency_type' is provided.

    """

    # Compute inlet state
    state_in = fluid.set_state(props.HmassP_INPUTS, h_in, p_in)
    state_out_is = fluid.set_state(props.PSmass_INPUTS, p_out, state_in.s)

    # Evaluate compression process
    if efficiency_type == "isentropic":
        # Compute outlet state according to the definition of isentropic efficiency
        h_out = state_in.h + (state_out_is.h - state_in.h) / efficiency
        state_out = fluid.set_state(props.HmassP_INPUTS, h_out, p_out)
        states = [state_in, state_out]

    elif efficiency_type == "polytropic":
        # Differential equation defining the polytropic compression
        def odefun(p, h):
            state = fluid.set_state(props.HmassP_INPUTS, h, p)
            dhdp = 1.0 / (efficiency * state.rho)
            return dhdp, state

        # Solve polytropic compression differential equation
        sol = solve_ivp(
            lambda p, h: odefun(p, h)[0],
            [state_in.p, p_out],
            [state_in.h],
            t_eval=np.linspace(state_in.p, p_out, num_steps),
            method="RK45",
        )

        # Evaluate fluid properties at intermediate states
        states = postprocess_ode(sol.t, sol.y, odefun)
        state_in, state_out = states[0], states[-1]

    else:
        raise ValueError("Invalid efficiency_type. Use 'isentropic' or 'polytropic'.")

    # Compute work
    isentropic_work = state_out_is.h - state_in.h
    specific_work = state_out.h - state_in.h

    # Compute degree of superheating
    state_in = props.calculate_superheating(state_in, fluid)
    state_out = props.calculate_superheating(state_out, fluid)

    # Compute degree of subcooling
    state_in = props.calculate_subcooling(state_in, fluid)
    state_out = props.calculate_subcooling(state_out, fluid)

    # Create result dictionary
    result = {
        "type": "compressor",
        "fluid_name": fluid.name,
        "states": props.states_to_dict(states),
        "state_in": state_in,
        "state_out": state_out,
        "efficiency": efficiency,
        "efficiency_type": efficiency_type,
        "specific_work": specific_work,
        "isentropic_work": isentropic_work,
        "pressure_ratio": state_out.p / state_in.p,
        "mass_flow": np.nan,
        "color": "black",
    }

    return result


def expansion_process(
    fluid, h_in, p_in, p_out, efficiency, efficiency_type="isentropic", num_steps=50
):
    """
    Calculate properties along a compression process defined by a isentropic or polytropic efficiency

    Parameters
    ----------
    fluid : Fluid
        The fluid object used to evaluate thermodynamic properties
    h_in : float
        Enthalpy at the start of the compression process.
    p_in : float
        Pressure at the start of the compression process.
    p_out : float
        Pressure at the end of the compression process.
    efficiency : float
        The efficiency of the compression process.
    efficiency_type : str, optional
        The type of efficiency to be used in the process ('isentropic' or 'polytropic'). Default is 'isentropic'.
    num_steps : int, optional
        The number of steps for the polytropic process calculation. Default is 50.

    Returns
    -------
    tuple
        Tuple containing (state_out, states) where states is a list with all intermediate states

    Raises
    ------
    ValueError
        If an invalid 'efficiency_type' is provided.

    """

    # Compute inlet state
    state_in = fluid.set_state(props.HmassP_INPUTS, h_in, p_in)
    state_out_is = fluid.set_state(props.PSmass_INPUTS, p_out, state_in.s)
    if efficiency_type == "isentropic":
        # Compute outlet state according to the definition of isentropic efficiency
        h_out = state_in.h - efficiency * (state_in.h - state_out_is.h)
        state_out = fluid.set_state(props.HmassP_INPUTS, h_out, p_out)
        states = [state_in, state_out]

    elif efficiency_type == "polytropic":
        # Differential equation defining the polytropic expansion
        def odefun(p, h):
            state = fluid.set_state(props.HmassP_INPUTS, h, p)
            dhdp = efficiency / state.rho
            return dhdp, state

        # Solve polytropic expansion differential equation
        sol = solve_ivp(
            lambda p, h: odefun(p, h)[0],
            [state_in.p, p_out],
            [state_in.h],
            t_eval=np.linspace(state_in.p, p_out, num_steps),
            method="RK45",
        )

        # Evaluate fluid properties at intermediate states
        states = postprocess_ode(sol.t, sol.y, odefun)
        state_in, state_out = states[0], states[-1]

    else:
        raise ValueError("Invalid efficiency_type. Use 'isentropic' or 'polytropic'.")

    # Compute work
    isentropic_work = state_in.h - state_out_is.h
    specific_work = state_in.h - state_out.h

    # Compute degree of superheating
    state_in = props.calculate_superheating(state_in, fluid)
    state_out = props.calculate_superheating(state_out, fluid)

    # Compute degree of subcooling
    state_in = props.calculate_subcooling(state_in, fluid)
    state_out = props.calculate_subcooling(state_out, fluid)

    # Create result dictionary
    result = {
        "type": "turbine",
        "fluid_name": fluid.name,
        "states": props.states_to_dict(states),
        "state_in": state_in,
        "state_out": state_out,
        "efficiency": efficiency,
        "efficiency_type": efficiency_type,
        "specific_work": specific_work,
        "isentropic_work": isentropic_work,
        "pressure_ratio": state_in.p / state_out.p,
        "mass_flow": np.nan,
        "color": "black",
    }

    return result


def isenthalpic_valve(state_in, p_out, fluid, N=50):
    p_array = np.linspace(state_in.p, p_out, N)
    h_array = state_in.h * p_array
    states = []
    for p, h in zip(p_array, h_array):
        states.append(fluid.set_state(props.HmassP_INPUTS, h, p))
    return states


def postprocess_ode(t, y, ode_handle):
    # Collect additional properties for each integration step
    ode_out = []
    for t_i, y_i in zip(t, y.T):
        _, state = ode_handle(t_i, y_i)
        ode_out.append(state)
    return ode_out



from pydantic import (
    BaseModel,
    ValidationError,
    Field,
    model_validator,
    field_validator,
    ConfigDict,
)
from typing import Optional, List, Union, Literal, Dict
from enum import Enum
from typing_extensions import Self
from typing_extensions import Annotated
from pydantic.functional_validators import AfterValidator, BeforeValidator
from ..axial_turbine import (
    VARIABLES,
    LOSS_MODELS,
    LOSS_COEFFICIENTS,
    CHOKING_MODELS,
    DEVIATION_MODELS,
    OBJECTIVE_FUNCTIONS,
    CONSTRAINTS,
    RADIUS_TYPE,
)
from ..pysolver_view import (
    SOLVER_OPTIONS,
    VALID_LIBRARIES_AND_METHODS,
    DERIVATIVE_METHODS,
)

DEFAULT_VARIABLES = {
    "v_in": {"value": 0.1, "lower_bound": 0.001, "upper_bound": 0.5},
    "w_out": {
        "value": [0.65, 0.65],
        "lower_bound": [0.1, 0.1],
        "upper_bound": [1.0, 1.0],
    },
    "s_out": {
        "value": [0.15, 0.15],
        "lower_bound": [0.0, 0.0],
        "upper_bound": [0.32, 0.32],
    },
    "beta_out": {
        "value": [0.83, 0.17],
        "lower_bound": [0.72, 0.06],
        "upper_bound": [0.94, 0.28],
    },
    "v*_in": {
        "value": [0.4, 0.4],
        "lower_bound": [0.1, 0.1],
        "upper_bound": [1.0, 1.0],
    },
    "beta*_throat": {
        "value": [0.83, 0.17],
        "lower_bound": [0.72, 0.06],
        "upper_bound": [0.94, 0.28],
    },
    "w*_throat": {
        "value": [0.65, 0.65],
        "lower_bound": [0.1, 0.1],
        "upper_bound": [1.0, 1.0],
    },
    "s*_throat": {
        "value": [0.15, 0.15],
        "lower_bound": [0.0, 0.0],
        "upper_bound": [0.32, 0.32],
    },
    "specific_speed": {"value": 1.2, "lower_bound": 0.01, "upper_bound": 10},
    "blade_jet_ratio": {"value": 0.5, "lower_bound": 0.1, "upper_bound": 0.9},
    "hub_tip_ratio_in": {
        "value": [0.6, 0.6],
        "lower_bound": [0.6, 0.6],
        "upper_bound": [0.9, 0.9],
    },
    "hub_tip_ratio_out": {
        "value": [0.6, 0.6],
        "lower_bound": [0.6, 0.6],
        "upper_bound": [0.9, 0.9],
    },
    "aspect_ratio": {
        "value": [1.5, 1.5],
        "lower_bound": [1.0, 1.0],
        "upper_bound": [2.0, 2.0],
    },
    "pitch_chord_ratio": {
        "value": [0.9, 0.9],
        "lower_bound": [0.75, 0.75],
        "upper_bound": [1.10, 1.10],
    },
    "trailing_edge_thickness_opening_ratio": {
        "value": [0.1, 0.1],
        "lower_bound": [0.05, 0.05],
        "upper_bound": [0.4, 0.4],
    },
    "leading_edge_angle": {
        "value": [0.41, 0.5],
        "lower_bound": [0.41, 0.08],
        "upper_bound": [0.92, 0.58],
    },
    "gauging_angle": {
        "value": [0.17, 0.94],
        "lower_bound": [0.06, 0.72],
        "upper_bound": [0.28, 0.94],
    },
    "throat_location_fraction": {
        "value": [1.0, 1.0],
        "lower_bound": None,
        "upper_bound": None,
    },
    "leading_edge_diameter": {
        "value": [2 * 0.00600448, 2 * 0.00600448],
        "lower_bound": None,
        "upper_bound": None,
    },
    "leading_edge_wedge_angle": {
        "value": [45.00, 45.00],
        "lower_bound": None,
        "upper_bound": None,
    },
    "tip_clearance": {"value": [0.00, 5e-4], "lower_bound": None, "upper_bound": None},
    "cascade_type": {
        "value": ["stator", "rotor"],
        "lower_bound": None,
        "upper_bound": None,
    },
}
"""
Default set of variables for design optimization.
"""


def check_string(input: str, input_type, options) -> str:
    """
    Check if a string is within a list of options.

    This function checks whether the provided string is present in the list of options.
    It raises an AssertionError if the string is not found in the options list.

    Parameters
    ----------
    input : str
        The string to check.
    input_type : str
        The type of the input, used for error message formatting.
    options : list
        List of valid options.

    Returns
    -------
    str
        The input string if it is found in the options list.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """
    assert (
        input in options
    ), f"Invalid {input_type}: '{input}'. Available options: {options}"
    return input


class OperationPoints(BaseModel):
    """
    Model describing the operational points of an axial turbine.

    Attributes
    ----------
    fluid_name : str
        Name of the fluid.
    T0_in : float
        Inlet total temperature (must be greater than 0).
    p0_in : float
        Inlet total pressure (must be greater than 0).
    p_out : float
        Outlet pressure (must be greater than 0).
    omega : float
        Rotational speed (must be greater than 0).
    alpha_in : float
        Inlet angle.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    fluid_name: str
    T0_in: Annotated[float, Field(gt=0)]
    p0_in: Annotated[float, Field(gt=0)]
    p_out: Annotated[float, Field(gt=0)]
    omega: Annotated[float, Field(gt=0)]
    alpha_in: float


class PerformanceMap(BaseModel):
    """
    Model describing a performance map for axial turbine analysis.

    Attributes
    ----------
    fluid_name : Union[str, List[str]]
        Name of the fluid or list of fluid names.
    T0_in : float, list
        Inlet total temperature or list of inlet total temperatures (each must be greater than 0).
    p0_in : float, list
        Inlet total pressure or list of inlet total pressures (each must be greater than 0).
    p_out : float, list
        Outlet pressure or list of outlet pressures (each must be greater than 0).
    omega : float, list
        Rotational speed or list of rotational speeds (each must be greater than 0).
    alpha_in : float, list
        Inlet angle or list of inlet angles.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    fluid_name: Union[str, List[str]]
    T0_in: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    p0_in: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    p_out: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    omega: Union[Annotated[float, Field(gt=0)], List[Annotated[float, Field(gt=0)]]]
    alpha_in: Union[float, List[float]]


class TuningFactors(BaseModel):
    """
    Model describing tuning factors of loss coefficients.

    Attributes
    ----------
    profile : float, optional
        Tuning factor for profile loss coefficient. Default is 1.00.
    incidence : float, optional
        Tuning factor for incidence loss coefficient. Default is 1.00.
    secondary : float, optional
        Tuning factor for secondary loss coefficient. Default is 1.00.
    trailing : float, optional
        Tuning factor for trailing loss coefficient. Default is 1.00.
    clearance : float, optional
        Tuning factor for clearance loss coefficient. Default is 1.00.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    profile: float = 1.00
    incidence: float = 1.00
    secondary: float = 1.00
    trailing: float = 1.00
    clearance: float = 1.00

LossModelEnum = Enum('LossModels', dict(zip([model.upper() for model in LOSS_MODELS], LOSS_MODELS)))
LossCoefficientEnum = Enum('LossCoefficients', dict(zip([model.upper() for model in LOSS_COEFFICIENTS], LOSS_COEFFICIENTS)))

class LossModel(BaseModel):
    """
    Model describing loss model setup and related parameters.

    Attributes
    ----------
    model : str
        The type of loss model. Should be one of the predefined loss models.
    loss_coefficient : str
        The loss coefficient type. Should be one of the predefined loss coefficients.
    custom_value : float
        Loss coefficient value if model is 'custom'
    inlet_displacement_thickness_height_ratio : float, optional
        Ratio of inlet displacement thickness to height. Default is 0.011.
    tuning_factors : TuningFactors, optional
        Object containing tuning factors for loss coefficients. Default is a new instance of TuningFactors.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    model : LossModelEnum
    loss_coefficient : LossCoefficientEnum
    custom_value: Annotated[float, Field(ge=0)] = 0.0
    inlet_displacement_thickness_height_ratio: float = 0.011
    tuning_factors: TuningFactors = TuningFactors()

DeviationModelEnum = Enum('DeviationModels', dict(zip([model.upper() for model in DEVIATION_MODELS], DEVIATION_MODELS)))
ChokingModelEnum = Enum('ChokingModels', dict(zip([model.upper() for model in CHOKING_MODELS], CHOKING_MODELS)))

class SimulationOptions(BaseModel):
    """
    Model describing simulation options.

    Attributes
    ----------
    deviation_model : str
        The type of deviation model. Should be one of the predefined deviation models.
    blockage_model : str, float, optional
        The blockage model. Can be a string or a float value. Default is 0.00.
    choking_model : str
        The choking model. Should be one of the predefined choking models.
    rel_step_fd : float, optional
        Relative step for finite difference calculations (only for evaluate_cascade_critical choking model). Default is 1e-4.
    loss_model : LossModel
        Object describing loss model setup and related parameters.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    deviation_model : DeviationModelEnum
    blockage_model: Union[str, float] = 0.00
    choking_model : ChokingModelEnum
    rel_step_fd: float = 1e-4
    loss_model: LossModel

SolverOptionEnum = Enum('SolverOptions', dict(zip([model.upper() for model in SOLVER_OPTIONS], SOLVER_OPTIONS)))
DerivativeMethodEnum = Enum('DerivativeMethods', dict(zip([model.upper() for model in DERIVATIVE_METHODS], DERIVATIVE_METHODS)))

class SolverOptionsPerformanceAnalysis(BaseModel):
    """
    Model describing solver options for performance analysis.

    Attributes
    ----------
    method : str, optional
        The root solver method. Should be one of the predefined solver options. Default is 'hybr'.
    tolerance : float, optional
        Tolerance for convergence. Default is 1e-8.
    max_iterations : int, optional
        Maximum number of iterations. Default is 100.
    derivative_method : str, optional
        The method for finite difference approximation calculation. Should be one of the predefined derivative methods. Default is '2-point'.
    derivative_abs_step : float, optional
        Absolute step size for finite difference approximation calculation. Default is 1e-6.
    print_convergence : bool, optional
        Flag indicating whether to print convergence information. Default is True.
    plot_convergence : bool, optional
        Flag indicating whether to plot convergence. Default is False.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    method : SolverOptionEnum = SOLVER_OPTIONS[0]
    tolerance: float = 1e-8
    max_iterations: int = 100
    derivative_method : DerivativeMethodEnum = DERIVATIVE_METHODS[0]
    derivative_abs_step: float = 1e-6
    print_convergence: bool = True
    plot_convergence: bool = False

CascadeTypeEnum = Enum('CascadeTypes', dict(zip([model.upper() for model in ["stator", "rotor"]], ["stator", "rotor"])))

class GeometryPerformanceAnalysis(BaseModel):
    """
    Model describing geometry parameters for performance analysis.

    Attributes
    ----------
    cascade_type : list[str]
        Types of cascades, either 'stator' or 'rotor'.
    radius_hub_in : list[float]
        Hub radius of the cascades inlet (each must be greater than 0).
    radius_hub_out : list[float]
        Hub radius of the cascades exit (each must be greater than 0).
    radius_tip_in : list[float]
        Tip radius of the cascades inlet (each must be greater than 0).
    radius_tip_out : list[float]
        Tip radius of the cascades exit (each must be greater than 0).
    pitch : list[float]
        Pitch of the cascades (each must be greater than 0).
    chord : list[float]
        Chord of the cascades (each must be greater than 0).
    stagger_angle : list[float]
        Stagger angle of the cascades, in degrees (each must be between -90 and 90).
    opening : list[float]
        Throat opening of the cascades (each must be greater than 0).
    leading_edge_angle : list[float]
        Leading edge angle of the cascades, in degrees (each must be between -90 and 90).
    leading_edge_wedge_angle : list[float]
        Leading edge wedge angle of the cascades, in degrees (each must be between 0 and 90).
    leading_edge_diameter : list[float]
        Diameter of the leading edge of the cascades (each must be greater than 0).
    trailing_edge_thickness : list[float]
        Thickness of the trailing edge of the cascades (each must be greater than 0).
    maximum_thickness : list[float]
        Maximum thickness of the cascades (each must be greater than 0).
    tip_clearance : list[float]
        Tip clearance of the cascades (each must be greater than or equal 0).
    throat_location_fraction : list[float]
        Fractional location of the throat in the cascades (each must be between 0 and 1).

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    cascade_type: List[CascadeTypeEnum]
    radius_hub_in: List[Annotated[float, Field(gt=0)]]
    radius_hub_out: List[Annotated[float, Field(gt=0)]]
    radius_tip_in: List[Annotated[float, Field(gt=0)]]
    radius_tip_out: List[Annotated[float, Field(gt=0)]]
    pitch: List[Annotated[float, Field(gt=0)]]
    chord: List[Annotated[float, Field(gt=0)]]
    stagger_angle: List[Annotated[float, Field(le=90), Field(gt=-90)]]
    opening: List[Annotated[float, Field(gt=0)]]
    leading_edge_angle: List[Annotated[float, Field(le=90), Field(gt=-90)]]
    leading_edge_wedge_angle: List[Annotated[float, Field(le=90), Field(gt=0)]]
    leading_edge_diameter: List[Annotated[float, Field(gt=0)]]
    trailing_edge_thickness: List[Annotated[float, Field(gt=0)]]
    maximum_thickness: List[Annotated[float, Field(gt=0)]]
    tip_clearance: List[Annotated[float, Field(ge=0)]]
    throat_location_fraction: List[Annotated[float, Field(le=1), Field(gt=0)]]

    @model_validator(mode="after")
    def check_length(self) -> Self:
        attributes = vars(self).values()
        if not all(len(attr) == len(self.cascade_type) for attr in attributes):
            raise ValueError("Geometry input is not of same length")
        return self

LibraryEnum = Enum('LibraryOptions', dict(zip([model.upper() for model in VALID_LIBRARIES_AND_METHODS.keys()], VALID_LIBRARIES_AND_METHODS.keys())))
MethodEnum = Enum('MethodOption', dict(zip([model.upper() for model in [x for xs in VALID_LIBRARIES_AND_METHODS.values() for x in xs]], [x for xs in VALID_LIBRARIES_AND_METHODS.values() for x in xs])))
UpdateOnEnum = Enum('UpdateOns', dict(zip([model.upper() for model in ["gradient", "function"]], ["gradient", "function"])))

class SolverOptionsOptimization(BaseModel):
    """
    Model describing solver options for optimization.

    Checks that the given solver method corresponds with the given library.

    Attributes
    ----------
    library : str, optional
        The optimization library to be used. Should be one of the predefined libraries. Default is 'scipy'.
    method : str, optional
        The optimization method to be used. Default is 'slsqp'.
    tolerance : float, optional
        Tolerance for convergence. Default is 1e-5.
    max_iterations : int, optional
        Maximum number of iterations. Default is 100.
    derivative_method : str, optional
        The method for finite difference calculation. Should be one of the predefined derivative methods. Default is '2-point'.
    derivative_abs_step : float, optional
        Absolute step size for finite difference calculation. Default is None.
    print_convergence : bool, optional
        Flag indicating whether to print convergence information. Default is True.
    plot_convergence : bool, optional
        Flag indicating whether to plot convergence. Default is False.
    update_on : str, optional
        The criteria for updating. Should be either 'gradient' or 'function'. Default is 'gradient'.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    library: LibraryEnum = list(VALID_LIBRARIES_AND_METHODS.keys())[0]
    method: MethodEnum = VALID_LIBRARIES_AND_METHODS[list(VALID_LIBRARIES_AND_METHODS.keys())[0]][0]
    tolerance: float = 1e-5
    max_iterations: int = 100
    derivative_method: DerivativeMethodEnum = DERIVATIVE_METHODS[0]
    derivative_abs_step: float = None
    print_convergence: bool = True
    plot_convergence: bool = False
    update_on: UpdateOnEnum = "gradient"

    @model_validator(mode="after")
    def check_solver_method(self) -> Self:
        lib = self.library
        method = self.method
        if not method in VALID_LIBRARIES_AND_METHODS[lib]:
            raise ValueError(f"Method {method} is not available in {lib} library.")
        return self


class PerformanceAnalysis(BaseModel):
    """
    Model describing performance analysis parameters.

    Attributes
    ----------
    performance_map : PerformanceMap, optional
        Object containing performance map data.
    solver_options : SolverOptionsPerformanceAnalysis, optional
        Object containing solver options for performance analysis.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    performance_map: PerformanceMap = None
    solver_options: SolverOptionsPerformanceAnalysis = (
        SolverOptionsPerformanceAnalysis()
    )


class Variable(BaseModel):
    """
    Model describing a variable.

    If the attributes are lists, each value corresponds to a seperate cascade.

    Checks that the type and length of the attributes are the same.

    Attributes
    ----------
    value : float, str, list[float], list[str]
        The value(s) of the variable, which could be a float, string, list of floats, or list of strings.
    lower_bound : float, list[float], optional
        The lower bound of the variable. Default is None.
    upper_bound : float, list[float], optional
        The upper bound of the variable. Default is None.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    value : Union[float, CascadeTypeEnum, List[float], List[CascadeTypeEnum]]
    lower_bound: Union[float, List[float]] = None
    upper_bound: Union[float, List[float]] = None

    @model_validator(mode="after")
    def check_length(self) -> Self:
        attributes = vars(self).values()
        input_type = type(self.value)
        if self.lower_bound is not None:
            if not all(isinstance(attr, input_type) for attr in attributes):
                raise ValueError("Variable input is not of same type")
            if isinstance(self.value, list):
                if not all(len(attr) == len(self.value) for attr in attributes):
                    raise ValueError("Variable input is not of same length")
        return self


class Constraint(BaseModel):
    """
    Model describing a constraint.

    Attributes
    ----------
    type : str
        The type of constraint, indicating whether it is less than ('<'), equal to ('='), or greater than ('>').
    value : float
        The value of the constraint.
    normalize : bool
        Flag indicating whether the constraint should be normalized.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    type: Literal["<", "=", ">"]
    value: float
    normalize: bool

ObjectiveFunctionEnum = Enum('ObjectiveFunctions', dict(zip([model.upper() for model in OBJECTIVE_FUNCTIONS], OBJECTIVE_FUNCTIONS)))
VariableEnum = Enum('Variables', dict(zip([model.upper() for model in VARIABLES], VARIABLES)))
ConstraintEnum = Enum('Constraints', dict(zip([model.upper() for model in CONSTRAINTS], CONSTRAINTS)))
RadiusTypeEnum = Enum('RadiusTypes', dict(zip([model.upper() for model in RADIUS_TYPE], RADIUS_TYPE)))

class DesignOptimization(BaseModel):
    """
    Model describing design optimization parameters.

    Checks that variables contain all variables from a predifned list (VARIABLES). Returns a default dictionary if not.

    Attributes
    ----------
    objective_function : str, optional
        The objective function for optimization. Should be one of the predefined objective functions. Default is 'efficiency_ts'.
    variables : dict[str], optional
        Dictionary containing variables for optimization. Should contain a dict for each variable in a predefined variable list (VARIABLES).
    constraints : dict[str], optional
        Dictionary containing constraints for optimization.
    solver_options : SolverOptionsOptimization, optional
        Object containing solver options for optimization.
    radius_type : str, optional
        The type of radius used in optimization. Should be one of the predefined radius types. Default is 'constant_mean'.

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid", use_enum_values=True)
    objective_function: ObjectiveFunctionEnum = OBJECTIVE_FUNCTIONS[0]
    variables: Dict[
        VariableEnum,
        Variable,
    ] = None
    constraints: Dict[
        ConstraintEnum,
        Constraint,
    ] = None
    solver_options: SolverOptionsOptimization = SolverOptionsOptimization()
    radius_type : RadiusTypeEnum = RADIUS_TYPE[0]

    # After validator to check variables holds all variables
    @field_validator("variables")
    def check_variables(cls, v: Dict) -> Dict:
        if v == None:
            # Give completely default turbine
            v = DEFAULT_VARIABLES
            print(
                "Design optimization variables are not provided. Defualt set of variables are used."
            )
        lacking_keys = set(VARIABLES) - set(v.keys())
        if len(lacking_keys) > 0:
            # Fill in lacking keys
            v = DEFAULT_VARIABLES
            print(
                f"Design optimization variables are missing: {lacking_keys}. Defualt set of variables are used."
            )
        return v


class AxialTurbine(BaseModel):
    """
    Model describing an axial turbine.

    Attributes
    ----------
    turbomachinery : str
        The type of turbomachinery, which for this class is "axial_turbine".
    operation_points : OperationPoints
        Object containing operation points data.
    simulation_options : SimulationOptions
        Object containing simulation options data.
    geometry : GeometryPerformanceAnalysis, optional
        Object containing geometry performance analysis data. Default is None.
    performance_analysis : PerformanceAnalysis, optional
        Object containing performance analysis data. Default is a new instance of PerformanceAnalysis().
    design_optimization : DesignOptimization, optional
        Object containing design optimization data. Default is a new instance of DesignOptimization().

    Configurations
    --------------
    extra : str, optional
        Indicates that no extra input is allowed. Default is "forbid".
    """

    model_config = ConfigDict(extra="forbid")
    turbomachinery: Literal["axial_turbine"]
    operation_points: Union[List[OperationPoints], OperationPoints]
    simulation_options: SimulationOptions
    geometry: GeometryPerformanceAnalysis = None
    performance_analysis: PerformanceAnalysis = PerformanceAnalysis()
    design_optimization: DesignOptimization = DesignOptimization()



import numpy as np
from .. import math


def compute_losses(input_parameters):
    r"""

    Evaluate loss coefficient according to :cite:`kacker_mean_1982` and :cite:`moustapha_improved_1990`.

    This model split the total loss coefficient into profile, secondary, trailing edge, tip clearance and incidence.
    The model adopt the incidence loss coefficient proposed by  :cite:`moustapha_improved_1990`, but is otherwise similar to the
    model proposed by :cite:`kacker_mean_1982`.
    The loss coefficient are combined through the following relation:

    .. math::

        \mathrm{Y_{tot}} = \mathrm{Y_{p}} + \mathrm{Y_{te}} + \mathrm{Y_{corr}}\mathrm{Y_{s}} + \mathrm{Y_{cl}} + \mathrm{Y_{inc}}

    The function calls a function for each loss component:

        - :math:`\mathrm{Y_{p}}` : `get_profile_loss`
        - :math:`\mathrm{Y_{te}}` : `get_trailing_edge_loss`
        - :math:`\mathrm{Y_{s}}` : `get_secondary_loss`
        - :math:`\mathrm{Y_{cl}}` : `get_tip_clearance_loss`
        - :math:`\mathrm{Y_{inc}}` : `get_incidence_loss`
        - :math:`\mathrm{Y_{corr}}` : `get_secondary_loss_correction_factor`

    Parameters
    ----------
    input_parameters : dict
        A dictionary containing containing the keys, `flow`, `geometry` and `options`.

    Returns
    -------
    dict
        A dictionary with the loss components.

    """

    # Load data
    flow_parameters = input_parameters["flow"]
    geometry = input_parameters["geometry"]
    beta_des = geometry["leading_edge_angle"]

    # Profile loss coefficient
    Y_p = get_profile_loss(flow_parameters, geometry)

    # Secondary loss coefficient
    Y_s = get_secondary_loss(flow_parameters, geometry)

    # Tip clearance loss coefficient
    Y_cl = get_tip_clearance_loss(flow_parameters, geometry)

    # Trailing edge loss coefficienct
    Y_te = get_trailing_edge_loss(flow_parameters, geometry)

    # Incidence loss for profile loss
    Y_inc = get_incidence_loss(flow_parameters, geometry, beta_des)

    # Incidence correction factor for secondary loss coefficient
    Y_corr = get_secondary_loss_correction_factor(flow_parameters, geometry)
    Y_s *= Y_corr

    # Calculate total pressure loss coefficient
    Y = Y_p + Y_s + Y_cl + Y_te + Y_inc

    # Return a dictionary of loss components
    losses = {
        "loss_profile": Y_p,
        "loss_incidence": Y_inc,
        "loss_trailing": Y_te,
        "loss_secondary": Y_s,
        "loss_clearance": Y_cl,
        "loss_total": Y_p + Y_te + Y_inc + Y_s + Y_cl,
    }

    return losses


def get_profile_loss(flow_parameters, geometry):
    r"""
    Calculate the profile loss coefficient for the current cascade using the Kacker and Okapuu loss model.
    The equation for :math:`\mathrm{Y_p}` is given by:

    .. math::

        \mathrm{Y_p} = \mathrm{Y_{reaction}} - \left|\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right| \cdot
        \left(\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right) \cdot (\mathrm{Y_{impulse}} - \mathrm{Y_{reaction}})

    where:

        - :math:`\mathrm{Y_{reaction}}` is the reaction loss coefficient computed using Aungier correlation.
        - :math:`\mathrm{Y_{impulse}}` is the impulse loss coefficient computed using Aungier correlation.
        - :math:`\theta_\mathrm{in}` is the inlet metal angle.
        - :math:`\beta_\mathrm{out}` is the exit flow angle.

    The function also applies various corrections based on flow parameters and geometry factors.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Profile loss coefficient.

    """

    # Load data
    Re = flow_parameters["Re_out"]
    Ma_rel_out = flow_parameters["Ma_rel_out"]
    Ma_rel_in = flow_parameters["Ma_rel_in"]
    p0rel_in = flow_parameters["p0_rel_in"]
    p_in = flow_parameters["p_in"]
    p0rel_out = flow_parameters["p0_rel_out"]
    p_out = flow_parameters["p_out"]
    beta_out = flow_parameters["beta_out"]

    r_ht_in = geometry["hub_tip_ratio_in"]
    s = geometry["pitch"]
    c = geometry["chord"]
    theta_in = geometry["leading_edge_angle"]
    t_max = geometry["maximum_thickness"]
    cascade_type = geometry["cascade_type"]

    # Reynolds number correction factor
    f_Re = (
        (Re / 2e5) ** (-0.4) * (Re < 2e5)
        + 1 * (Re >= 2e5 and Re <= 1e6)
        + (Re / 1e6) ** (-0.2) * (Re > 1e6)
    )

    # Mach number correction factor
    f_Ma = 1 + 60 * (Ma_rel_out - 1) ** 2 * (Ma_rel_out > 1)

    # Compute losses related to shock effects at the inlet of the cascade
    f_hub = get_hub_to_mean_mach_ratio(r_ht_in, cascade_type)
    a = max(0, f_hub * Ma_rel_in - 0.4)
    Y_shock = 0.75 * a**1.75 * r_ht_in * (p0rel_in - p_in) / (p0rel_out - p_out)
    Y_shock = max(0, Y_shock)

    # Compute compressible flow correction factors
    Kp, K2, K1 = get_compressible_correction_factors(Ma_rel_in, Ma_rel_out)

    # Yp_reaction and Yp_impulse according to Aungier correlation
    # These formulas are valid for 40<abs(angle_out)<80
    # Extrapolating outside of this limits might give completely wrong results
    # If the optimization algorithm has upper and lower bounds for the outlet
    # angle there is no need to worry about this problem
    # angle_out_bis keeps the 40deg-losses for outlet angles lower than 40deg
    angle_out_bis = max(abs(beta_out), 40)
    Yp_reaction = nozzle_blades(s / c, angle_out_bis)
    Yp_impulse = impulse_blades(s / c, angle_out_bis)

    # Formula according to Kacker-Okapuu
    Y_p = Yp_reaction - abs(theta_in / beta_out) * (theta_in / beta_out) * (
        Yp_impulse - Yp_reaction
    )

    # Limit the extrapolation of the profile loss to avoid negative values for
    # blade profiles with little deflection
    # Low limit to 80% of the axial entry nozzle profile loss
    # This value is completely arbitrary
    Y_p = max(Y_p, 0.8 * Yp_reaction)

    # Avoid unphysical effect on the thickness by defining the variable aa
    aa = max(0, -theta_in / beta_out)
    Y_p = Y_p * ((t_max / c) / 0.2) ** aa
    Y_p = 0.914 * (2 / 3 * Y_p * Kp + Y_shock)

    # Corrected profile loss coefficient
    Y_p = f_Re * f_Ma * Y_p

    return Y_p


def get_secondary_loss(flow_parameters, geometry):
    r"""
    The function calculates the secondary loss coefficient using the Kacker-Okapuu model.
    The main equation for :math:`\mathrm{Y_s}` is given by:

    .. math::

        \mathrm{Y_s} = 1.2 \cdot \mathrm{K_s} \cdot 0.0334 \cdot far \cdot Z \cdot \frac{\cos(\beta_{out})}{\cos(\theta_{in})}

    where:

        - :math:`K_s` is a correction factor accounting for compressible flow effects.
        - :math:`far` is a factor that account for the aspect ratio of the current cascade.
        - :math:`Z` is a blade loading parameter.
        - :math:`\beta_\mathrm{out}` is the exit flow angle.
        - :math:`\theta_\mathrm{in}` is the inlet metal angle.

    The correction factor :math:`\mathrm{K_s}` from a separate correction factor given from `get_compressible_correction_factors`.
    `get_compressible_correction_factors` returns :math:`\mathrm{K_s}`, which is used to calculate :math:`\mathrm{K_s}`:

    .. math::

        \mathrm{K_s} = 1 - \left(\frac{b}{H}\right)^2 (1-\mathrm{K_p})

    where:

        - :math:`b` is the axial chord.
        - :math:`H` is the mean blade height.

    The aspect ratio factor is calculated from the following correlation:

    .. math:: 

        far = \begin{cases}
                1 - \frac{0.25 \sqrt{|2-H/c|}}{H/c} & \text{if } H/c < 2.0 \\
                c/H & \text{if } H/c \geq 2.0
                \end{cases}

    where :math:`c` is the blade chord.

    The loading parameters is caclualated from the following correlation:

    .. math::

        Z = 4 (\tan(\beta_\mathrm{in}) - \tan(\beta_\mathrm{out}))^2\frac{\cos^2(\beta_\mathrm{out})}{\cos(\beta_m)} 

    where:

        - :math:`\beta_\mathrm{in}` and :math:`\beta_\mathrm{out}` is the inlet and exit relative flow angle
        - :math:`\beta_m = \tan^{-1}(0.5(\tan(\beta_\mathrm{in})-\tan(\beta_\mathrm{out})))` is the mean gas angle.
           
    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.

    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Secondary loss coefficient.

    """

    Ma_rel_out = flow_parameters["Ma_rel_out"]
    Ma_rel_in = flow_parameters["Ma_rel_in"]
    beta_out = flow_parameters["beta_out"]
    beta_in = flow_parameters["beta_in"]

    b = geometry["axial_chord"]
    H = geometry["height"]
    c = geometry["chord"]
    theta_in = geometry["leading_edge_angle"]

    # Compute compressible flow correction factors
    Kp, K2, K1 = get_compressible_correction_factors(Ma_rel_in, Ma_rel_out)

    # Secondary loss coefficient
    K3 = (b / H) ** 2
    Ks = 1 - K3 * (1 - Kp)
    Ks = max(0.1, Ks)
    angle_m = math.arctand((math.tand(beta_in) + math.tand(beta_out)) / 2)
    Z = (
        4
        * (math.tand(beta_in) - math.tand(beta_out)) ** 2
        * math.cosd(beta_out) ** 2
        / math.cosd(angle_m)
    )
    far = (1 - 0.25 * np.sqrt(abs(2 - H / c))) / (H / c) * (H / c < 2) + 1 / (H / c) * (
        H / c >= 2
    )
    Y_s = 1.2 * Ks * 0.0334 * far * Z * math.cosd(beta_out) / math.cosd(theta_in)

    return Y_s


def get_trailing_edge_loss(flow_parameters, geometry):
    r"""
    Calculate the trailing edge loss coefficient using the Kacker-Okapuu model.
    The main equation for the kinetic-energy coefficient is given by:

    .. math::

        d_{\phi^2} = d_{\phi^2_\mathrm{reaction}} - \left|\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right| \cdot
        \left(\frac{\theta_\mathrm{in}}{\beta_\mathrm{out}}\right) \cdot (d_{\phi^2_\mathrm{impulse}} - d_{\phi^2_\mathrm{reaction}})

    The kinetic-energy coefficient is converted to the total pressure loss coefficient by:

    .. math::

        \mathrm{Y_{te}} = \frac{1}{{1 - \phi^2}} - 1

    where:

        - :math:`d_{\phi^2_\mathrm{reaction}}` and :math:`d_{\phi^2_\mathrm{impulse}}` are coefficients related to kinetic energy loss for reaction and impulse blades respectively, and are interpolated based on trailing edge to throat opening ratio.
        - :math:`\beta_\mathrm{out}` is the exit flow angle.
        - :math:`\theta_\mathrm{in}` is the inlet metal angle.

    The function also applies various interpolations and computations based on flow parameters and geometry factors.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Trailing edge loss coefficient.

    """

    t_te = geometry["trailing_edge_thickness"]
    o = geometry["opening"]
    angle_in = geometry["leading_edge_angle"]
    angle_out = flow_parameters["beta_out"]

    # Range of trailing edge to throat opening ratio
    r_to_data = [0, 0.2, 0.4]

    # Reacting blading
    phi_data_reaction = [0, 0.045, 0.15]

    # Impulse blading
    phi_data_impulse = [0, 0.025, 0.075]

    # Numerical trick to avoid too big r_to's
    r_to = min(0.4, t_te / o)

    # Interpolate data
    d_phi2_reaction = np.interp(r_to, r_to_data, phi_data_reaction)
    d_phi2_impulse = np.interp(r_to, r_to_data, phi_data_impulse)

    # Compute kinetic energy loss coefficient
    d_phi2 = d_phi2_reaction - abs(angle_in / angle_out) * (angle_in / angle_out) * (
        d_phi2_impulse - d_phi2_reaction
    )

    # Limit the extrapolation of the trailing edge loss
    d_phi2 = max(d_phi2, d_phi2_impulse / 2)
    Y_te = 1 / (1 - d_phi2) - 1

    return Y_te


def get_tip_clearance_loss(flow_parameters, geometry):
    r"""
    Calculate the tip clearance loss coefficient for the current cascade using the Kacker and Okapuu loss model.
    The equation for the tip clearance loss coefficent is given by:

    .. math::

        \mathrm{Y_{cl}} = B \cdot Z \cdot \frac{c}{H} \cdot \left(\frac{t_\mathrm{cl}}{H}\right)^{0.78}

    where:

        - :math:`B` is an empirical parameter that depends on the type of cascade (0 for stator, 0.37 for shrouded rotor).
        - :math:`Z` is a blade loading parameter
        - :math:`c` is the chord.
        - :math:`H` is the mean blade height.
        - :math:`t_\mathrm{cl}` is the tip clearance.


    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Tip clearance loss coefficient.
    """

    beta_out = flow_parameters["beta_out"]
    beta_in = flow_parameters["beta_in"]

    H = geometry["height"]
    c = geometry["chord"]
    t_cl = geometry["tip_clearance"]
    cascade_type = geometry["cascade_type"]

    # Calculate blade loading parameter Z
    angle_m = math.arctand((math.tand(beta_in) + math.tand(beta_out)) / 2)
    Z = (
        4
        * (math.tand(beta_in) - math.tand(beta_out)) ** 2
        * math.cosd(beta_out) ** 2
        / math.cosd(angle_m)
    )

    # Empirical parameter (0 for stator, 0.37 for shrouded rotor)
    if cascade_type == "stator":
        B = 0
    elif cascade_type == "rotor":
        B = 0.37
    else:
        print("Specify the type of cascade")

    # Tip clearance loss coefficient
    Y_cl = B * Z * c / H * (t_cl / H) ** 0.78

    return Y_cl


def get_incidence_loss(flow_parameters, geometry, beta_des):
    r"""
    Calculate the incidence loss coefficient according to the correlation proposed by :cite:`moustapha_improved_1990`.
    
    The model first computes the incidence parameter, :math:`\chi`. Based on this parameter, the kinetic-energy incidence 
    loss coefficient is calculated to match experimental data.

    .. math::

        \Delta\phi^2_p = \begin{cases}
                        -5.1734e^{-6}\chi + 7.6902e^{-9}\chi^2 & \text{if } -800 \leq \chi \leq 0 \\
                        0.778e^{-5}\chi + 0.56e^{-7}\chi^2 + 0.4e^{-10}\chi^3 + 2.054e^{-19}\chi^6 & \text{if } 0 \leq \chi \leq 800
                        \end{cases}

    The kinetic-energy coefficient is converted to total pressure loss coefficient through the `convert_kinetic_energy_coefficient` function. 

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.
    beta_des : float
        Inlet flow angle for zero incidence losses (design).

    Returns
    -------
    float
        Incidence loss coefficient.

    Raises
    ------
    Warning
        If incidence parameter (:math:`chi`) is out of range.

    """

    Ma_rel_out = flow_parameters["Ma_rel_out"]  # Exit relative mach number
    beta_in = flow_parameters["beta_in"]  # Inlet relativ flow angle
    gamma = flow_parameters["gamma_out"]  # Specific heat ratio

    s = geometry["pitch"]  # Pitch
    le = geometry["leading_edge_diameter"]  # Leading edge diameter
    theta_in = geometry["leading_edge_angle"]  # Inlet metal angle
    theta_out = geometry["gauging_angle"]  # Exit metal angle

    # Compute incidence parameter
    chi = get_incidence_parameter(le, s, theta_in, theta_out, beta_in, beta_des)

    # Check if incidence parameter is within the range of the experimental data
    if abs(chi) > 800:
        raise Warning("Incidence parameter out of range: chi = {chi}")

    # Calculate kinetic-energy loss coefficient
    if chi >= 0:
        dPhi = 0.778e-5 * chi + 0.56e-7 * chi**2 + 0.4e-10 * chi**3 + 2.054e-19 * chi**6
    elif chi < 0:
        dPhi = -5.1734e-6 * chi + 7.6902e-9 * chi**2

    # Convert kinetic-energy loss coefficient to total pressure loss coeffcient
    Y_inc = convert_kinetic_energy_coefficient(dPhi, gamma, Ma_rel_out)

    return Y_inc


def get_secondary_loss_correction_factor(flow_parameters, geometry):
    r"""
    Calculate the correction factor for the secondary loss coefficient to account for incidence losses.

    The correction factor is calculated based on the secondary flow incidence parameter (:math:`\chi`), which is determined using
    geometrical and flow parameters.

    The correction factor (:math:`Y_\mathrm{corr}`) is given by:

    .. math::

        Y_\mathrm{corr} = \begin{cases}
        e^{0.9\chi} + 13\chi^2 + 400\chi^4 & \text{if } \chi \geq 0 \\
        e^{0.9\chi} & \text{if } \chi < 0
        \end{cases}

    and the incidence paramter is calculated as:

    .. math::

        \chi = \frac{\beta_\mathrm{in} - \theta_\mathrm{in}}{180 - (\theta_\mathrm{in} + \theta_\mathrm{out})}\left(\frac{\cos(\theta_\mathrm{in})}{\cos(\theta_\mathrm{out})}\right)^{-1.5}\left(\frac{\mathrm{d_{le}}}{c}\right)^{-0.3}

    where:

        - :math:`\chi` is the secondary flow incidence parameter.
        - :math:`c` is the chord length.
        - :math:`\mathrm{d_{le}}` is the leading edge diameter.
        - :math:`\theta_\mathrm{in}` is the inlet metal angle.
        - :math:`\theta_\mathrm{out}` is the exit metal angle.
        - :math:`\beta_\mathrm{in}` is the inlet relative flow angle.

    The function also checks if the secondary incidence parameter falls within the range of the experimental data used to 
    determine the correction factor.

    Parameters
    ----------
    flow_parameters : dict
        Dictionary containing flow-related parameters.
    geometry : dict
        Dictionary with geometric parameters.

    Returns
    -------
    float
        Secondary loss correction factor.

    """

    beta_in = flow_parameters["beta_in"]

    c = geometry["chord"]
    le = geometry["leading_edge_diameter"]
    theta_in = geometry["leading_edge_angle"]
    theta_out = geometry["gauging_angle"]

    chi = (
        (abs(beta_in) - abs(theta_in))
        / (180 - (theta_in + theta_out))
        * (math.cosd(theta_in) / math.cosd(theta_out)) ** -1.5
        * (le / c) ** -0.3
    )

    if not (-0.4 < chi < 0.3):
        raise Warning("Secondary incidence parameter out of range: chi = {chi}")

    if chi >= 0:
        Y_corr = np.exp(0.9 * chi) + 13 * chi**2 + 400 * chi**4
    else:
        Y_corr = np.exp(0.9 * chi)

    return Y_corr


def nozzle_blades(r_sc, angle_out):
    r"""
    Use Aungier correlation to compute the pressure loss coefficient for nozzle blades :cite:`aungier_turbine_2006`.

    This correlation is a formula that reproduces the figures from the Ainley and Mathieson original figures :cite:`ainley_method_1951`,
    and is a function of the pitch-to-chord ratio and exit relative flow angle. 

    The correlation uses the following equations:

    .. math::

        & \beta_\mathrm{tan} = 90 - \beta_\mathrm{ax} \\
        & \left(\frac{s}{c}\right)_\mathrm{min} = \begin{cases}
                                                    0.46 + \frac{\beta_\mathrm{tan}}{77} && \text{if } \beta_\mathrm{tan} < 30 \\
                                                    0.614 + \frac{\beta_\mathrm{tan}}{130} && \text{if } \beta_\mathrm{tan} \geq 30
                                                \end{cases} \\
        & X = \left(\frac{s}{c}\right) - \left(\frac{s}{c}\right)_\mathrm{min} \\
        & A = \begin{cases}
                0.025 + \frac{27 - \beta_\mathrm{tan}}{530} && \text{if } \beta_\mathrm{tan} < 27 \\
                0.025 + \frac{27 - \beta_\mathrm{tan}}{3085} && \text{if } \beta_\mathrm{tan} \geq 27
              \end{cases} \\
        & B = 0.1583 - \frac{\beta_\mathrm{tan}}{1640} \\
        & C = 0.08\left(\frac{\beta_\mathrm{tan}}{30}\right)^2 -1 \\
        & n = 1 + \frac{\beta_\mathrm{tan}}{30} \\
        & \mathrm{Y_{p,reaction}} = \begin{cases}
                A + BX^2 + CX^3 && \text{if } \beta_\mathrm{tan} < 30 \\
                A + B |X|^n && \text{if } \beta_\mathrm{tan} \geq 30
              \end{cases}   

    where:

        - :math:`s` is the pitch
        - :math:`c` is the chord
        - :math:`\beta_\mathrm{tan}` and :math:`\beta_\mathrm{ax}` is the exit relative flow angle with respect to tangential and axial direction. 
        
    Parameters
    ----------
    r_sc : float
        Pitch-to-chord ratio.
    angle_out : float
        Exit relative flow angle (in degrees).

    Returns
    -------
    float
        Pressure loss coefficient for impulse blades.

    """

    phi = 90 - angle_out
    r_sc_min = (0.46 + phi / 77) * (phi < 30) + (0.614 + phi / 130) * (phi >= 30)
    X = r_sc - r_sc_min
    A = (0.025 + (27 - phi) / 530) * (phi < 27) + (0.025 + (27 - phi) / 3085) * (
        phi >= 27
    )
    B = 0.1583 - phi / 1640
    C = 0.08 * ((phi / 30) ** 2 - 1)
    n = 1 + phi / 30
    Yp_reaction = (A + B * X**2 + C * X**3) * (phi < 30) + (A + B * abs(X) ** n) * (
        phi >= 30
    )

    return Yp_reaction


def impulse_blades(r_sc, angle_out):
    r"""
    Use Aungier correlation to compute the pressure loss coefficient for impulse blades :cite:`aungier_turbine_2006`.

    This correlation is a formula that reproduces the figures from the Ainley and Mathieson original figures :cite:`ainley_method_1951`,
    and is a function of the pitch-to-chord ratio and exit relative flow angle. 

    The correlation uses the following equations:

    .. math::

        & \beta_\mathrm{tan} = 90 - \beta_\mathrm{ax} \\
        & \left(\frac{s}{c}\right)_\mathrm{min} = 0.224 + 1.575\left(\frac{\beta_\mathrm{tan}}{90}\right) - \left(\frac{\beta_\mathrm{tan}}{90}\right)^2 \\
        & X = \left(\frac{s}{c}\right) - \left(\frac{s}{c}\right)_\mathrm{min} \\
        & A = 0.242 - \frac{\beta_\mathrm{tan}}{151} + \left(\frac{\beta_\mathrm{tan}}{127}\right)^2 \\
        & B = \begin{cases}
                0.3 + \frac{30 - \beta_\mathrm{tan}}{50} && \text{if } \beta_\mathrm{tan} < 30 \\
                0.3 + \frac{30 - \beta_\mathrm{tan}}{275} && \text{if } \beta_\mathrm{tan} \geq 30
              \end{cases} \\
        & C = 0.88 - \frac{\beta_\mathrm{tan}}{42.4} + \left(\frac{\beta_\mathrm{tan}}{72.8}\right)^2 \\
        & \mathrm{Y_{p,impulse}} = A + BX^2 - CX^3 

    where:

        - :math:`s` is the pitch
        - :math:`c` is the chord
        - :math:`\beta_\mathrm{tan}` and :math:`\beta_\mathrm{ax}` is the exit relative flow angle with respect to tangential and axial direction. 
        
    Parameters
    ----------
    r_sc : float
        Pitch-to-chord ratio.
    angle_out : float
        Exit relative flow angle (in degrees).

    Returns
    -------
    float
        Pressure loss coefficient for impulse blades.

    """

    phi = 90 - angle_out
    r_sc_min = 0.224 + 1.575 * (phi / 90) - (phi / 90) ** 2
    X = r_sc - r_sc_min
    A = 0.242 - phi / 151 + (phi / 127) ** 2
    B = (0.3 + (30 - phi) / 50) * (phi < 30) + (0.3 + (30 - phi) / 275) * (phi >= 30)
    C = 0.88 - phi / 42.4 + (phi / 72.8) ** 2
    Yp_impulse = A + B * X**2 - C * X**3

    return Yp_impulse


def get_compressible_correction_factors(Ma_rel_in, Ma_rel_out):
    r"""

    Compute compressible flow correction factor according to Kacker and Okapuu loss model :cite:`kacker_mean_1982`.

    The correction factors :math:`\mathrm{K_1}`, :math:`\mathrm{K_2}` and :math:`\mathrm{K_p}` was introduced by :cite:`kacker_mean_1982` to correct previous correlation (:cite:`ainley_method_1951`)
    for effect of higher mach number and channel acceleration. The correction factors reduces the losses at higher mach number. Their definition follows: 

    .. math::

        &\mathrm{K_1} = \begin{cases}
                1 & \text{if } \mathrm{Ma_{out}} < 0.2 \\
                1 - 1.25(\mathrm{Ma_{out}} - 0.2) & \text{if } \mathrm{Ma_{out}} \geq 0.2
              \end{cases} \\
        &\mathrm{K_2} = \left(\frac{\mathrm{Ma_{in}}}{\mathrm{Ma_{out}}}\right) \\
        &\mathrm{K_p} = 1 - \mathrm{K_2}(1-\mathrm{K_1})

    where:

        - :math:`\mathrm{Ma_{in}}` is the relative mach number at the cascade inlet.
        -  :math:`\mathrm{Ma_{out}}` is the relative mach number at the cascade exit.
    
    Parameters
    ----------
    Ma_rel_in : float
        Inlet relative mach number.
    Ma_rel_out : float
        Exit relative mach number.

    Returns
    -------
    float
        Correction factor :math:`\mathrm{K_p}`.
    float
        Correction factor :math:`\mathrm{K_2}`.
    float
        Correction factor :math:`\mathrm{K_1}`.
    """

    K1 = 1 * (Ma_rel_out < 0.2) + (1 - 1.25 * (Ma_rel_out - 0.2)) * (
        Ma_rel_out > 0.2 and Ma_rel_out < 1.00
    )
    K2 = (Ma_rel_in / Ma_rel_out) ** 2
    Kp = 1 - K2 * (1 - K1)
    Kp = max(0.1, Kp)
    return [Kp, K2, K1]


def get_hub_to_mean_mach_ratio(r_ht, cascade_type):
    r"""
    Compute the ratio between Mach at the hub and mean span at the inlet of the current cascade.

    Due to radial variation in gas conditions, Mach at the hub will always be higher than at the mean.
    Thus, shock losses at the hub could occur even when the Mach is subsonic at the mean blade span.

    Parameters
    ----------
    r_ht : float
        Hub to tip ratio at the inlet of the current cascade.
    cascade_type : str
        Type of the current cascade, either 'stator' or 'rotor'.

    Returns
    -------
    float
        Ratio between Mach at the hub and mean span at the inlet of the current cascade.

    """

    if r_ht < 0.5:
        r_ht = 0.5  # Numerical trick to prevent extrapolation

    r_ht_data = [0.5, 0.6, 0.7, 0.8, 0.9, 1.0]

    # Stator curve
    f_data_S = [1.4, 1.18, 1.05, 1.0, 1.0, 1.0]

    # Rotor curve
    f_data_R = [2.15, 1.7, 1.35, 1.12, 1.0, 1.0]

    if cascade_type == "stator":
        f = np.interp(r_ht, r_ht_data, f_data_S)
    elif cascade_type == "rotor":
        f = np.interp(r_ht, r_ht_data, f_data_R)
    else:
        print("Specify the type of cascade")

    return f


def get_incidence_parameter(le, s, theta_in, theta_out, beta_in, beta_des):
    r"""
    Calculate the incidence parameter according to the correlation proposed by :cite:`moustapha_improved_1990`.

    The incidence parameter is used to calculate the increment in profile losses due to the effect of incidence according to function `get_incidence_loss`.

    The quantity is calculated as:

    .. math::

        \chi = \left(\frac{\mathrm{d_{le}}}{s}\right)^{-1.6}\left(\frac{\cos{\theta_\mathrm{in}}}{\cos{\theta_\mathrm{out}}}\right)^{-2}(\beta_\mathrm{in} - \beta_\mathrm{des})

    where:

        - :math:`\mathrm{d_{le}}` is the leading edge diameter.
        - :math:`s` is the pitch.
        - :math:`\theta_\mathrm{in}` and :math:`\theta_\mathrm{out}` is the blade metal angle at the inlet and outlet respectively.
        - :math:`\beta_\mathrm{in}` and :math:`\beta_\mathrm{des}` is the inlet relative flow angle at given and design conditions respectively.

    Parameters
    ----------
    le : float
        Leading edge diameter.
    s : float
        Pitch.
    theta_in : float
        Leading edge metal angle (in degrees).
    theta_out : float
        Trailing edge metal angle (in degrees).
    beta_in : float
        Inlet relative flow angle (in degrees).
    beta_des : float
        Inlet relative flow angle at design condition (in degrees).

    Returns
    -------
    float
        Incidence parameter.
    """

    # TODO: explain smoothing/blending tricks

    chi = (
        (le / s) ** (-1.6)
        * (math.cosd(theta_in) / math.cosd(theta_out)) ** (-2)
        * (abs(beta_in) - abs(beta_des))
    )
    return chi


def convert_kinetic_energy_coefficient(dPhi, gamma, Ma_rel_out):
    r"""

    Convert the kinetic energy coefficient increment due to incidence to the total pressure loss coefficient according to the following correlation:

    .. math::

        \mathrm{Y} = \frac{\left(1-\frac{\gamma -1}{2}\mathrm{Ma_{out}}^2(\frac{1}{(1-\Delta\phi^2_p)}-1)\right)^\frac{-\gamma}{\gamma - 1}-1}{1-\left(1 + \frac{\gamma - 1}{2}\mathrm{Ma_{out}}^2\right)^\frac{-\gamma}{\gamma - 1}}

    where:

        - :math:`\gamma` is the specific heat ratio.
        - :math:`\mathrm{Ma_{out}}` is the cascade exit relative mach number.
        - :math:`\Delta\phi^2_p` is the kinetic energy loss coefficient increment due to incidence.

    Parameters
    ----------
    dPhi : float
        Kinetic energy coefficient increment.
    gamma : float
        Heat capacity ratio.
    Ma_rel_out : float
        The cascade exit relative mach number.

    Returns
    -------
    float
        The total pressure loss coefficient.

    Warnings
    --------
    This conversion assumes that the fluid is a perfect gas.

    """

    denom = 1 - (1 + (gamma - 1) / 2 * Ma_rel_out**2) ** (-gamma / (gamma - 1))
    numer = (1 - (gamma - 1) / 2 * Ma_rel_out**2 * (1 / (1 - dPhi) - 1)) ** (
        -gamma / (gamma - 1)
    ) - 1

    Y = numer / denom

    return Y



# Import packages
import os
import sys


if __name__ == "__main__":

    # Define package name for re-usability
    package_name = "turboflow"

    # Define bash command to append cwd to PYTHONPATH
    # PACKAGE_PATH = os.getcwd()
    PACKAGE_PATH = os.path.dirname(os.getcwd())
    bashrc_header = f"# Append {package_name} package to PYTHONPATH"
    if sys.platform == "win32":  # Windows
        bashrc_line = f'export PYTHONPATH=$PYTHONPATH\;"{PACKAGE_PATH}"'
    else:  # Linux or MacOS
        bashrc_line = f'export PYTHONPATH=$PYTHONPATH:"{PACKAGE_PATH}"'

    # Locate the .bashrc file
    bashrc_path = os.path.expanduser("~/.bashrc")

    # Ask for user confirmation with default set to no
    response = input(
        f"Do you want to add the {package_name} path to your .bashrc? [yes/NO]: "
    )
    if response.lower() in ["y", "yes"]:
        try:
            # Check if the line already exists in the .bashrc
            with open(bashrc_path, "r") as file:
                if bashrc_line in file.read():
                    print(f".bashrc already contains the {package_name} path.")
                else:
                    with open(bashrc_path, "a") as file_append:
                        file_append.write(f"\n{bashrc_header}")
                        file_append.write(f"\n{bashrc_line}\n")
                    print(f"Path to {package_name} package added to .bashrc")
                    print(
                        f"Restart the terminal or run 'source ~/.bashrc' for the changes to take effect."
                    )
        except Exception as e:
            print(f"An error occurred: {e}")
    else:
        print("Operation aborted by user.")
